#========================================================================================
# package      DE_EPAGES::Saferpay2::t::unit::Tools.t
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Saferpay2::API::Tools
#========================================================================================
package DE_EPAGES::Saferpay2::t::unit::Tools;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Saferpay2::API::Tools qw ( HandleTransactionResponse );
use Test::More;

use strict;


sub test {
    my $self = shift;

    $self->test_HandleInitializeErrorResponse_WithError();
    $self->test_HandleInitializeErrorResponse_WithoutError();
    $self->test_HandleAssertErrorResponse_WithError();
    $self->test_HandleAssertErrorResponse_WithoutError();
    done_testing();
}


sub test_HandleInitializeErrorResponse_WithError {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hParam = {
        ErrorName => 'TRANSACTION_DECLINED'
    };

    my $hExpectedHandledResponse = {
        ErrorName => 'TRANSACTION_DECLINED',
        HasErrorsInResponse => 1,
        Transaction         => {
            Id => 'NONE',
            Status => 'FAILED',
            ExternalStatus => 'TRANSACTION_DECLINED',
        }
    };

    my $hHandledResponse = HandleTransactionResponse($hParam);
    is_deeply($hHandledResponse, $hExpectedHandledResponse, "$SubTestName - Response handled correctly");
}

sub test_HandleInitializeErrorResponse_WithoutError {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hParam = {};

    my $hExpectedHandledResponse = {
        HasErrorsInResponse => 0,
        Transaction         => {
            ExternalStatus => undef,
        }
    };

    my $hHandledResponse = HandleTransactionResponse($hParam);
    is_deeply($hHandledResponse, $hExpectedHandledResponse, "$SubTestName - Response handled correctly");
}


sub test_HandleAssertErrorResponse_WithError {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hParam = {
        ErrorName   => 'GENERAL_DECLINED',
        TransactionId => '123'
    };

    my $hExpectedHandledResponse = {
        ErrorName => 'GENERAL_DECLINED',
        HasErrorsInResponse => 1,
        TransactionId => '123',
        Transaction         => {
            Id => '123',
            Status => 'FAILED',
            ExternalStatus => 'GENERAL_DECLINED',
        }
    };

    my $hHandledResponse = HandleTransactionResponse($hParam);
    is_deeply($hHandledResponse, $hExpectedHandledResponse, "$SubTestName - Response handled correctly");
}


sub test_HandleAssertErrorResponse_WithoutError {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hParam = {
        Transaction => {
            Id     => '123',
            Status => 'AUTHORIZED'
        }
    };

    my $hExpectedHandledResponse = {
        HasErrorsInResponse => 0,
        Transaction         => {
            Id     => '123',
            Status => 'AUTHORIZED',
            ExternalStatus => 'AUTHORIZED',
        }
    };

    my $hHandledResponse = HandleTransactionResponse($hParam);
    is_deeply($hHandledResponse, $hExpectedHandledResponse, "$SubTestName - Response handled correctly");
}

__PACKAGE__->runScript();
